from typing import Any, Literal, TypedDict, get_args


Currency = Literal[
    "USD",
    "AUD",
    "BTC",
    "CHF",
    "ETH",
    "EUR",
    "GBP",
    "JPY",
]


CurrencySlug = Literal[
    "usd",
    "aud",
    "btc",
    "chf",
    "eth",
    "eur",
    "gbp",
    "jpy",
]


class CurrencyDefinition(TypedDict):
    id: Currency
    slug: CurrencySlug
    label: str


SLUG_BY_CURRENCY = dict(zip(get_args(Currency), get_args(CurrencySlug)))


CURRENCY_BY_SLUG = dict(zip(get_args(CurrencySlug), get_args(Currency)))


def get_currency(currency_or_slug: Currency | CurrencySlug) -> CurrencyDefinition:
    if is_supported_currency(currency_or_slug):
        return CURRENCIES[currency_or_slug]
    if is_supported_currency_slug(currency_or_slug):
        return CURRENCIES[CURRENCY_BY_SLUG[currency_or_slug]]
    raise ValueError(f"Invalid currency {currency_or_slug}")


def is_currency_identifier(value: Any) -> bool:
    return is_supported_currency(value) or is_supported_currency_slug(value)


def is_supported_currency(value: Any) -> bool:
    return isinstance(value, str) and value in get_args(Currency)


def is_supported_currency_slug(value: Any) -> bool:
    return isinstance(value, str) and value in get_args(CurrencySlug)


class ReleasesTypeVersion(TypedDict, total=False):
    sulu: str  # f"{Deployment}.{Version}"
    encore: str  # f"{Deployment}.{Version}"
    phoenix: str  # f"{Deployment}.{Version}"


class CurrenciesType(TypedDict):
    ETH: CurrencyDefinition
    BTC: CurrencyDefinition
    USD: CurrencyDefinition
    EUR: CurrencyDefinition
    CHF: CurrencyDefinition
    GBP: CurrencyDefinition
    AUD: CurrencyDefinition
    JPY: CurrencyDefinition


CURRENCIES = {
    "ETH": {
        "id": "ETH",
        "label": "Ether",
        "slug": "eth",
    },
    "BTC": {
        "id": "BTC",
        "label": "Bitcoin",
        "slug": "btc",
    },
    "USD": {
        "id": "USD",
        "label": "United States dollar",
        "slug": "usd",
    },
    "EUR": {
        "id": "EUR",
        "label": "Euro",
        "slug": "eur",
    },
    "CHF": {
        "id": "CHF",
        "label": "Swiss franc",
        "slug": "chf",
    },
    "GBP": {
        "id": "GBP",
        "label": "Pound sterling",
        "slug": "gbp",
    },
    "AUD": {
        "id": "AUD",
        "label": "Australian dollar",
        "slug": "aud",
    },
    "JPY": {
        "id": "JPY",
        "label": "Japanese yen",
        "slug": "jpy",
    },
}



from typing import Any, Literal, get_args, TypedDict
from web3.types import ChecksumAddress
from .contracts import is_version, Contracts, Version
from .assets import PrimitiveAsset, CompoundV2Asset, ERC4626Asset, AssetType
from .networks import Network
from .types import Release


Status = Literal["live", "pending", "deprecated"]


Deployment = Literal["arbitrum", "base", "ethereum", "polygon", "testnet"]


def is_deployment(value: Any) -> bool:
    return isinstance(value, str) and value in get_args(Deployment)


def is_release(value: Any) -> bool:
    if isinstance(value, str) and "." in value:
        deployment, version = value.split(".")

        if is_deployment(deployment) and is_version(version):
            return version in RELEASES[deployment]

    return False


class ReleaseDefinition(TypedDict):
    # The unique release identifier.
    slug: str  # f"{Deployment}.{Version}"
    # The network identifier.
    network: Network
    # The address of the fund deployer contract.
    address: ChecksumAddress
    # The version (e.g. sulu, encore, phoenix) of the release.
    version: Version
    # The block number at which the fund deployer contract was deployed.
    inception: int
    # Mapping of contract names and their addresses within this release.
    contracts: Contracts
    # The release status (pending, deprecated or live).
    status: Status


Kind = Literal["test", "live"]


class DeploymentNamedTokensArbitrum(TypedDict):
    bal: ChecksumAddress
    comp: ChecksumAddress
    crv: ChecksumAddress
    cvx: ChecksumAddress
    dai: ChecksumAddress
    grt: ChecksumAddress
    mln: ChecksumAddress
    usdt: ChecksumAddress
    weth: ChecksumAddress


class DeploymentNamedTokensBase(TypedDict):
    comp: ChecksumAddress
    dai: ChecksumAddress
    mln: ChecksumAddress
    usdt: ChecksumAddress
    weth: ChecksumAddress


class DeploymentNamedTokensEthereum(TypedDict):
    aave: ChecksumAddress
    bal: ChecksumAddress
    ceth: ChecksumAddress
    comp: ChecksumAddress
    crv: ChecksumAddress
    cvx: ChecksumAddress
    dai: ChecksumAddress
    diva: ChecksumAddress
    ethx: ChecksumAddress
    grt: ChecksumAddress
    idle: ChecksumAddress
    lusd: ChecksumAddress
    mln: ChecksumAddress
    mpl: ChecksumAddress
    paxg: ChecksumAddress
    ptkn_mln: ChecksumAddress
    steth: ChecksumAddress
    sthoundeth: ChecksumAddress
    stkaave: ChecksumAddress
    stusd: ChecksumAddress
    sweth: ChecksumAddress
    uni: ChecksumAddress
    usda: ChecksumAddress
    usdc: ChecksumAddress
    usdt: ChecksumAddress
    weth: ChecksumAddress


class DeploymentNamedTokensPolygon(TypedDict):
    aave: ChecksumAddress
    bal: ChecksumAddress
    comp: ChecksumAddress
    crv: ChecksumAddress
    cvx: ChecksumAddress
    dai: ChecksumAddress
    eure: ChecksumAddress
    grt: ChecksumAddress
    mln: ChecksumAddress
    uni: ChecksumAddress
    usdc: ChecksumAddress
    usdt: ChecksumAddress
    weth: ChecksumAddress


DeploymentNamedTokens = (
    DeploymentNamedTokensArbitrum
    | DeploymentNamedTokensBase
    | DeploymentNamedTokensEthereum
    | DeploymentNamedTokensPolygon
)


class DeploymentNamedTokensAssetsArbitrum(TypedDict):
    bal: PrimitiveAsset
    comp: PrimitiveAsset
    crv: PrimitiveAsset
    cvx: PrimitiveAsset
    dai: PrimitiveAsset
    grt: PrimitiveAsset
    mln: PrimitiveAsset
    native_token_wrapper: PrimitiveAsset
    usdt: PrimitiveAsset
    weth: PrimitiveAsset


class DeploymentNamedTokensAssetsBase(TypedDict):
    comp: PrimitiveAsset
    dai: PrimitiveAsset
    mln: PrimitiveAsset
    native_token_wrapper: PrimitiveAsset
    usdt: PrimitiveAsset
    weth: PrimitiveAsset


class DeploymentNamedTokensAssetsEthereum(TypedDict):
    aave: PrimitiveAsset
    bal: PrimitiveAsset
    ceth: CompoundV2Asset
    comp: PrimitiveAsset
    crv: PrimitiveAsset
    cvx: PrimitiveAsset
    dai: PrimitiveAsset
    diva: PrimitiveAsset
    ethx: PrimitiveAsset
    grt: PrimitiveAsset
    idle: PrimitiveAsset
    lusd: PrimitiveAsset
    mln: PrimitiveAsset
    mpl: PrimitiveAsset
    native_token_wrapper: PrimitiveAsset
    paxg: PrimitiveAsset
    ptkn_mln: PrimitiveAsset
    sthoundeth: PrimitiveAsset
    stkaave: PrimitiveAsset
    steth: PrimitiveAsset
    stusd: ERC4626Asset
    sweth: PrimitiveAsset
    uni: PrimitiveAsset
    usda: PrimitiveAsset
    usdc: PrimitiveAsset
    usdt: PrimitiveAsset
    weth: PrimitiveAsset


class DeploymentNamedTokensAssetsPolygon(TypedDict):
    aave: PrimitiveAsset
    bal: PrimitiveAsset
    comp: PrimitiveAsset
    crv: PrimitiveAsset
    cvx: PrimitiveAsset
    dai: PrimitiveAsset
    eure: PrimitiveAsset
    grt: PrimitiveAsset
    mln: PrimitiveAsset
    native_token_wrapper: PrimitiveAsset
    uni: PrimitiveAsset
    usdc: PrimitiveAsset
    usdt: PrimitiveAsset
    weth: PrimitiveAsset


DeploymentNamedTokensAssets = (
    DeploymentNamedTokensAssetsArbitrum
    | DeploymentNamedTokensAssetsBase
    | DeploymentNamedTokensAssetsEthereum
    | DeploymentNamedTokensAssetsPolygon
)


class CoreSubgraphMapping(TypedDict):
    slug: str
    id: str
    devVersion: str


class AssetsSubgraphMapping(TypedDict):
    slug: str
    id: str


class SharesSubgraphMapping(TypedDict):
    slug: str
    id: str


class BalancesSubgraphMapping(TypedDict):
    slug: str
    id: str


class VaultsSubgraphMapping(TypedDict):
    slug: str
    id: str


class SubgraphMapping(TypedDict):
    core: CoreSubgraphMapping
    assets: AssetsSubgraphMapping
    shares: SharesSubgraphMapping
    balances: BalancesSubgraphMapping
    vaults: VaultsSubgraphMapping


class KnownAddressListIdMapping(TypedDict, total=False):
    noslippage_adapters: int
    adapters: int
    fees: int
    policies: int
    kiln_staking_contracts: int
    non_standard_price_feed_assets: int
    a_tokens: int
    deposit_wrapper_allowed_exchanges: int


class KnownAddressListIdMappingEthereumSpecific(KnownAddressListIdMapping):
    kiln_staking_contracts: int
    zero_lend_rwa_stablecoins_a_tokens: int
    zero_lend_lrt_btc_a_tokens: int


class KnownUintListIdMapping(TypedDict, total=False):
    allowed_morpho_blue_vaults: int


class ExternalContractsMapping(TypedDict):
    aaveUIIncentiveDataProvider: ChecksumAddress
    aaveV2IncentivesController: ChecksumAddress
    aaveV2LendingPoolProvider: ChecksumAddress
    aaveV3LendingPoolProvider: ChecksumAddress
    aaveV3ProtocolDataProvider: ChecksumAddress
    aaveV3RewardsController: ChecksumAddress
    aliceOrderManager: ChecksumAddress
    arrakisV2Helper: ChecksumAddress
    arrakisV2Resolver: ChecksumAddress
    balancerGaugeController: ChecksumAddress
    balancerHelpers: ChecksumAddress
    balancerMinter: ChecksumAddress
    balancerProtocolFeesCollector: ChecksumAddress
    balancerVault: ChecksumAddress
    chainlinkFeedsRegistry: ChecksumAddress
    compoundComptroller: ChecksumAddress
    compoundV3Rewards: ChecksumAddress
    curveChildLiquidityGaugeFactory: ChecksumAddress
    curveMinter: ChecksumAddress
    curveRegistry: ChecksumAddress
    cvxCrvStaking: ChecksumAddress
    gmxV2ChainlinkPriceFeed: ChecksumAddress
    gmxV2DataStore: ChecksumAddress
    gmxV2ExchangeRouter: ChecksumAddress
    gmxV2Reader: ChecksumAddress
    gmxV2ReferralStorage: ChecksumAddress
    kilnStaking: ChecksumAddress
    lidoWithdrawalsQueue: ChecksumAddress
    liquityCollSurplusPool: ChecksumAddress
    liquityHintHelpers: ChecksumAddress
    liquitySortedTroves: ChecksumAddress
    liquityTroveManager: ChecksumAddress
    makerMCDPotAddress: ChecksumAddress
    morphoBlue: ChecksumAddress
    multicall: ChecksumAddress
    paraswapV5AugustusSwapper: ChecksumAddress
    paraswapV5TokenTransferProxy: ChecksumAddress
    pendlePtLpOracle: ChecksumAddress
    staderStakingPoolManager: ChecksumAddress
    staderUserWithdrawManager: ChecksumAddress
    stakeWiseV3KeeperRewards: ChecksumAddress
    theGraphDelegationStakingProxy: ChecksumAddress
    theGraphEpochManagerProxy: ChecksumAddress
    uniswapV3NonFungiblePositionManager: ChecksumAddress
    voteLockedConvexToken: ChecksumAddress
    votiumVoteProxy: ChecksumAddress
    zeroExExchangeProxy: ChecksumAddress
    zeroExV4Exchange: ChecksumAddress
    zeroLendAaveV3UIIncentiveDataProvider: ChecksumAddress
    zeroLendLRTBTCAaveV3LendingPoolProvider: ChecksumAddress
    zeroLendLRTBTCAaveV3ProtocolDataProvider: ChecksumAddress
    zeroLendLRTBTCAaveV3RewardsController: ChecksumAddress
    zeroLendRWAStablecoinsAaveV3LendingPoolProvider: ChecksumAddress
    zeroLendRWAStablecoinsAaveV3ProtocolDataProvider: ChecksumAddress
    zeroLendRWAStablecoinsAaveV3RewardsController: ChecksumAddress


class DeploymentDefinition(TypedDict):
    slug: Deployment
    network: Network
    external_contracts: ExternalContractsMapping
    known_address_lists: (
        KnownAddressListIdMapping | KnownAddressListIdMappingEthereumSpecific
    )
    known_uint_lists: KnownUintListIdMapping
    kind: Kind
    address: ChecksumAddress
    label: str
    inception: int
    assets: list[AssetType]
    named_tokens: DeploymentNamedTokens
    subgraphs: SubgraphMapping
    releases: dict[Version, ReleaseDefinition]


# Release = Literal[
#     "arbitrum.sulu",
#     "base.sulu",
#     "ethereum.sulu",
#     "ethereum.encore",
#     "ethereum.phoenix",
#     "polygon.sulu",
#     "testnet.sulu",
# ] -> defined in types.py to avoid circular import


class ReleasesTypeVersion(TypedDict, total=False):
    sulu: str  # f"{Deployment}.{Version}"
    encore: str  # f"{Deployment}.{Version}"
    phoenix: str  # f"{Deployment}.{Version}"


class ReleasesType(TypedDict):
    arbitrum: ReleasesTypeVersion
    base: ReleasesTypeVersion
    ethereum: ReleasesTypeVersion
    polygon: ReleasesTypeVersion
    testnet: ReleasesTypeVersion


RELEASES: ReleasesType = {
    "arbitrum": {
        "sulu": "arbitrum.sulu",
    },
    "base": {
        "sulu": "base.sulu",
    },
    "ethereum": {
        "sulu": "ethereum.sulu",
        "encore": "ethereum.encore",
        "phoenix": "ethereum.phoenix",
    },
    "polygon": {
        "sulu": "polygon.sulu",
    },
    "testnet": {
        "sulu": "testnet.sulu",
    },
}



from typing import TypedDict
from web3.types import ChecksumAddress, TxParams
from .utils.clients import PublicClient, WalletClient
from ..abis import abis


# --------------------------------------------------------------------------------------------
# TRANSACTIONS
# --------------------------------------------------------------------------------------------


class SetFreelyTransferableSharesParams(TypedDict):
    client: WalletClient
    vault_proxy: ChecksumAddress


async def set_freely_transferable_shares(
    client: WalletClient,
    vault_proxy: ChecksumAddress,
) -> TxParams:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.setFreelyTransferableShares()
    return await client.populated_transaction(function)


class SetNominatedOwnerParams(TypedDict):
    client: WalletClient
    vault_proxy: ChecksumAddress
    next_nominated_owner: ChecksumAddress


async def set_nominated_owner(
    client: WalletClient,
    vault_proxy: ChecksumAddress,
    next_nominated_owner: ChecksumAddress,
) -> TxParams:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.setNominatedOwner(next_nominated_owner)
    return await client.populated_transaction(function)


class RemoveNominatedOwnerParams(TypedDict):
    client: WalletClient
    vault_proxy: ChecksumAddress


async def remove_nominated_owner(
    client: WalletClient,
    vault_proxy: ChecksumAddress,
) -> TxParams:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.removeNominatedOwner()
    return await client.populated_transaction(function)


class ClaimOwnershipParams(TypedDict):
    client: WalletClient
    vault_proxy: ChecksumAddress


async def claim_ownership(
    client: WalletClient,
    vault_proxy: ChecksumAddress,
) -> TxParams:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.claimOwnership()
    return await client.populated_transaction(function)


class SetNameParams(TypedDict):
    client: WalletClient
    vault_proxy: ChecksumAddress
    name: str


async def set_name(
    client: WalletClient,
    vault_proxy: ChecksumAddress,
    name: str,
) -> TxParams:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.setName(name)
    return await client.populated_transaction(function)


class SetSymbolParams(TypedDict):
    client: WalletClient
    vault_proxy: ChecksumAddress
    symbol: str


async def set_symbol(
    client: WalletClient,
    vault_proxy: ChecksumAddress,
    symbol: str,
) -> TxParams:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.setSymbol(symbol)
    return await client.populated_transaction(function)


async def add_asset_managers(
    client: WalletClient,
    vault_proxy: ChecksumAddress,
    managers: list[ChecksumAddress],
) -> TxParams:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.addAssetManagers(managers)
    return await client.populated_transaction(function)


async def remove_asset_managers(
    client: WalletClient,
    vault_proxy: ChecksumAddress,
    managers: list[ChecksumAddress],
) -> TxParams:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.removeAssetManagers(managers)
    return await client.populated_transaction(function)


# --------------------------------------------------------------------------------------------
# READ FUNCTIONS
# --------------------------------------------------------------------------------------------


async def get_name(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
) -> str:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.name()
    return await function.call()


async def get_symbol(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
) -> str:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.symbol()
    return await function.call()


async def get_owner(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.getOwner()
    return await function.call()


async def get_nominated_owner(
    client: PublicClient,
    vault: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(vault, abis.IVaultLib)
    function = contract.functions.getNominatedOwner()
    return await function.call()


async def get_denomination_asset(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.getDenominationAsset()
    return await function.call()


async def get_comptroller_proxy(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.getAccessor()
    return await function.call()


async def get_policy_manager(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.getPolicyManager()
    return await function.call()


async def get_fee_manager(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.getFeeManager()
    return await function.call()


async def shares_are_freely_transferable(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
) -> bool:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.sharesAreFreelyTransferable()
    return await function.call()


async def get_fund_deployer(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.getFundDeployer()
    return await function.call()



from typing import TypedDict, Any, Literal, get_args
from web3.types import ChecksumAddress


Version = Literal["phoenix", "encore", "sulu"]


def is_version(version: Any) -> bool:
    return isinstance(version, str) and version in get_args(Version)


class CommonContracts(TypedDict):
    ComptrollerLib: ChecksumAddress
    Dispatcher: ChecksumAddress
    FeeManager: ChecksumAddress
    FundDataProviderRouter: ChecksumAddress
    FundDeployer: ChecksumAddress
    FundValueCalculatorRouter: ChecksumAddress
    IntegrationManager: ChecksumAddress
    PolicyManager: ChecksumAddress
    VaultLib: ChecksumAddress


class SuluContracts(CommonContracts):
    AaveDebtPositionLib: ChecksumAddress
    AaveDebtPositionParser: ChecksumAddress
    AaveV3DebtPositionLib: ChecksumAddress
    AaveV3DebtPositionParser: ChecksumAddress
    AaveV3FlashLoanAssetManagerFactory: ChecksumAddress
    AaveV3FlashLoanAssetManagerLib: ChecksumAddress
    AavePriceFeed: ChecksumAddress
    AaveV2Adapter: ChecksumAddress
    AaveV2ATokenListOwner: ChecksumAddress
    AaveV3Adapter: ChecksumAddress
    AaveV3ATokenListOwner: ChecksumAddress
    AddressListRegistry: ChecksumAddress
    AlicePositionLib: ChecksumAddress
    AlicePositionParser: ChecksumAddress
    AllowedAdapterIncomingAssetsPolicy: ChecksumAddress
    AllowedAdaptersPerManagerPolicy: ChecksumAddress
    AllowedAdaptersPolicy: ChecksumAddress
    AllowedAssetsForRedemptionPolicy: ChecksumAddress
    AllowedDepositRecipientsPolicy: ChecksumAddress
    AllowedExternalPositionTypesPerManagerPolicy: ChecksumAddress
    AllowedExternalPositionTypesPolicy: ChecksumAddress
    AllowedRedeemersForSpecificAssetsPolicy: ChecksumAddress
    AllowedSharesTransferRecipientsPolicy: ChecksumAddress
    ArbitraryLoanPositionLib: ChecksumAddress
    ArbitraryLoanPositionParser: ChecksumAddress
    ArbitraryLoanTotalNominalDeltaOracleModule: ChecksumAddress
    ArrakisV2Adapter: ChecksumAddress
    ArrakisV2PriceFeed: ChecksumAddress
    AssetValueCalculator: ChecksumAddress
    BalancerV2GaugeTokenPriceFeed: ChecksumAddress
    BalancerV2LiquidityAdapter: ChecksumAddress
    BalancerV2StablePoolPriceFeed: ChecksumAddress
    BalancerV2WeightedPoolPriceFeed: ChecksumAddress
    ChainlinkLikeWstethPriceFeed: ChecksumAddress
    ChainlinkLikeYnEthPriceFeed: ChecksumAddress
    CompoundAdapter: ChecksumAddress
    CompoundDebtPositionLib: ChecksumAddress
    CompoundDebtPositionParser: ChecksumAddress
    CompoundPriceFeed: ChecksumAddress
    CompoundV3TokenListOwner: ChecksumAddress
    CompoundV3Adapter: ChecksumAddress
    ConvertedQuoteAggregatorFactory: ChecksumAddress
    ConvexVotingPositionLib: ChecksumAddress
    ConvexVotingPositionParser: ChecksumAddress
    CumulativeSlippageTolerancePolicy: ChecksumAddress
    CurveExchangeAdapter: ChecksumAddress
    CurveLiquidityAaveAdapter: ChecksumAddress
    CurveLiquidityAdapter: ChecksumAddress
    CurveLiquiditySethAdapter: ChecksumAddress
    CurveLiquidityStethAdapter: ChecksumAddress
    CurvePriceFeed: ChecksumAddress
    DepositWrapper: ChecksumAddress
    DisallowedAdapterIncomingAssetsPolicy: ChecksumAddress
    EntranceRateBurnFee: ChecksumAddress
    EntranceRateDirectFee: ChecksumAddress
    ERC4626Adapter: ChecksumAddress
    ERC4626PriceFeed: ChecksumAddress
    EtherFiEthPriceFeed: ChecksumAddress
    ExitRateBurnFee: ChecksumAddress
    ExitRateDirectFee: ChecksumAddress
    ExternalPositionFactory: ChecksumAddress
    ExternalPositionManager: ChecksumAddress
    FiduPriceFeed: ChecksumAddress
    FundValueCalculator: ChecksumAddress
    GasRelayPaymasterFactory: ChecksumAddress
    GasRelayPaymasterLib: ChecksumAddress
    GatedRedemptionQueueSharesWrapperFactory: ChecksumAddress
    GatedRedemptionQueueSharesWrapperLib: ChecksumAddress
    GenericAdapter: ChecksumAddress
    GlobalConfigLib: ChecksumAddress
    GlobalConfigProxy: ChecksumAddress
    GMXV2LeverageTradingPositionLib: ChecksumAddress
    GMXV2LeverageTradingPositionParser: ChecksumAddress
    KilnStakingPositionLib: ChecksumAddress
    KilnStakingPositionParser: ChecksumAddress
    LidoWithdrawalsPositionLib: ChecksumAddress
    LidoWithdrawalsPositionParser: ChecksumAddress
    LiquityDebtPositionLib: ChecksumAddress
    LiquityDebtPositionParser: ChecksumAddress
    ManagementFee: ChecksumAddress
    ManualValueOracleFactory: ChecksumAddress
    MapleLiquidityPositionLib: ChecksumAddress
    MapleLiquidityPositionParser: ChecksumAddress
    MinAssetBalancesPostRedemptionPolicy: ChecksumAddress
    MinMaxInvestmentPolicy: ChecksumAddress
    MinSharesSupplyFee: ChecksumAddress
    MorphoBluePositionLib: ChecksumAddress
    MorphoBluePositionParser: ChecksumAddress
    NoDepegOnRedeemSharesForSpecificAssetsPolicy: ChecksumAddress
    NotionalV2PositionLib: ChecksumAddress
    NotionalV2PositionParser: ChecksumAddress
    OneInchV5Adapter: ChecksumAddress
    OnlyRemoveDustExternalPositionPolicy: ChecksumAddress
    OnlyUntrackDustOrPricelessAssetsPolicy: ChecksumAddress
    ParaSwapV5Adapter: ChecksumAddress
    PeggedDerivativesPriceFeed: ChecksumAddress
    PendleV2Adapter: ChecksumAddress
    PendleV2PositionLib: ChecksumAddress
    PendleV2PositionParser: ChecksumAddress
    PendleMarketsRegistry: ChecksumAddress
    PerformanceFee: ChecksumAddress
    PoolTogetherV4Adapter: ChecksumAddress
    PoolTogetherV4PriceFeed: ChecksumAddress
    ProtocolFeeReserveLib: ChecksumAddress
    ProtocolFeeReserveProxy: ChecksumAddress
    ProtocolFeeTracker: ChecksumAddress
    SharePriceThrottledAssetManagerLib: ChecksumAddress
    SharePriceThrottledAssetManagerFactory: ChecksumAddress
    SharesSplitterFactory: ChecksumAddress
    SingleAssetRedemptionQueueLib: ChecksumAddress
    SingleAssetRedemptionQueueFactory: ChecksumAddress
    SolvV2BondBuyerPositionLib: ChecksumAddress
    SolvV2BondBuyerPositionParser: ChecksumAddress
    SolvV2BondIssuerPositionLib: ChecksumAddress
    SolvV2BondIssuerPositionParser: ChecksumAddress
    StaderSDPriceFeed: ChecksumAddress
    StaderStakingAdapter: ChecksumAddress
    StaderWithdrawalsPositionLib: ChecksumAddress
    StaderWithdrawalsPositionParser: ChecksumAddress
    StakeWiseV3StakingPositionLib: ChecksumAddress
    StakeWiseV3StakingPositionParser: ChecksumAddress
    SwellStakingAdapter: ChecksumAddress
    SynthetixAdapter: ChecksumAddress
    TermFinanceV1LendingPositionLib: ChecksumAddress
    TermFinanceV1LendingPositionParser: ChecksumAddress
    TheGraphDelegationPositionLib: ChecksumAddress
    TheGraphDelegationPositionParser: ChecksumAddress
    ThreeOneThirdAdapter: ChecksumAddress
    TransferAssetsAdapter: ChecksumAddress
    UintListRegistry: ChecksumAddress
    UniswapV2ExchangeAdapter: ChecksumAddress
    UniswapV2LiquidityAdapter: ChecksumAddress
    UniswapV2PoolPriceFeed: ChecksumAddress
    UniswapV3Adapter: ChecksumAddress
    UniswapV3LiquidityPositionLib: ChecksumAddress
    UniswapV3LiquidityPositionParser: ChecksumAddress
    UnpermissionedActionsWrapper: ChecksumAddress
    UsdEthSimulatedAggregator: ChecksumAddress
    ValueInterpreter: ChecksumAddress
    WstethPriceFeed: ChecksumAddress
    YearnVaultV2Adapter: ChecksumAddress
    YearnVaultV2PriceFeed: ChecksumAddress
    ZeroExV2Adapter: ChecksumAddress
    ZeroExV4Adapter: ChecksumAddress
    ZeroExV4AdapterPmm2Kyc: ChecksumAddress
    ZeroLendLRTBTCAaveV3Adapter: ChecksumAddress
    ZeroLendLRTBTCAaveV3ATokenListOwner: ChecksumAddress
    ZeroLendLRTBTCAaveV3DebtPositionLib: ChecksumAddress
    ZeroLendLRTBTCAaveV3DebtPositionParser: ChecksumAddress
    ZeroLendRWAStablecoinsAaveV3Adapter: ChecksumAddress
    ZeroLendRWAStablecoinsAaveV3ATokenListOwner: ChecksumAddress
    ZeroLendRWAStablecoinsAaveV3DebtPositionLib: ChecksumAddress
    ZeroLendRWAStablecoinsAaveV3DebtPositionParser: ChecksumAddress


class EncoreContracts(CommonContracts):
    AaveAdapter: ChecksumAddress
    AavePriceFeed: ChecksumAddress
    AdapterBlacklist: ChecksumAddress
    AdapterWhitelist: ChecksumAddress
    AggregatedDerivativePriceFeed: ChecksumAddress
    AlphaHomoraV1Adapter: ChecksumAddress
    AlphaHomoraV1PriceFeed: ChecksumAddress
    AssetBlacklist: ChecksumAddress
    AssetValueCalculator: ChecksumAddress
    AssetWhitelist: ChecksumAddress
    BuySharesCallerWhitelist: ChecksumAddress
    ChainlinkPriceFeed: ChecksumAddress
    CompoundAdapter: ChecksumAddress
    CompoundPriceFeed: ChecksumAddress
    CurveExchangeAdapter: ChecksumAddress
    CurveLiquidityAaveAdapter: ChecksumAddress
    CurveLiquidityEursAdapter: ChecksumAddress
    CurveLiquiditySethAdapter: ChecksumAddress
    CurveLiquidityStethAdapter: ChecksumAddress
    CurvePriceFeed: ChecksumAddress
    EntranceRateBurnFee: ChecksumAddress
    EntranceRateDirectFee: ChecksumAddress
    FundActionsWrapper: ChecksumAddress
    FundValueCalculator: ChecksumAddress
    GuaranteedRedemption: ChecksumAddress
    IdleAdapter: ChecksumAddress
    IdlePriceFeed: ChecksumAddress
    InvestorWhitelist: ChecksumAddress
    KyberAdapter: ChecksumAddress
    LidoStethPriceFeed: ChecksumAddress
    ManagementFee: ChecksumAddress
    MaxConcentration: ChecksumAddress
    MinMaxInvestment: ChecksumAddress
    ParaSwapV4Adapter: ChecksumAddress
    PerformanceFee: ChecksumAddress
    StakehoundEthPriceFeed: ChecksumAddress
    SynthetixAdapter: ChecksumAddress
    SynthetixPriceFeed: ChecksumAddress
    TrackedAssetsAdapter: ChecksumAddress
    UniswapV2Adapter: ChecksumAddress
    UniswapV2PoolPriceFeed: ChecksumAddress
    ValueInterpreter: ChecksumAddress
    WdgldPriceFeed: ChecksumAddress
    YearnVaultV2Adapter: ChecksumAddress
    YearnVaultV2PriceFeed: ChecksumAddress
    ZeroExV2Adapter: ChecksumAddress


class PhoenixContracts(CommonContracts):
    AaveAdapter: ChecksumAddress
    AavePriceFeed: ChecksumAddress
    AdapterBlacklist: ChecksumAddress
    AdapterWhitelist: ChecksumAddress
    AggregatedDerivativePriceFeed: ChecksumAddress
    AlphaHomoraV1Adapter: ChecksumAddress
    AlphaHomoraV1PriceFeed: ChecksumAddress
    AssetBlacklist: ChecksumAddress
    AssetValueCalculator: ChecksumAddress
    AssetWhitelist: ChecksumAddress
    BuySharesCallerWhitelist: ChecksumAddress
    ChaiAdapter: ChecksumAddress
    ChainlinkPriceFeed: ChecksumAddress
    ChaiPriceFeed: ChecksumAddress
    CompoundAdapter: ChecksumAddress
    CompoundPriceFeed: ChecksumAddress
    CurveExchangeAdapter: ChecksumAddress
    CurveLiquidityAaveAdapter: ChecksumAddress
    CurveLiquiditySethAdapter: ChecksumAddress
    CurveLiquidityStethAdapter: ChecksumAddress
    CurvePriceFeed: ChecksumAddress
    EntranceRateBurnFee: ChecksumAddress
    EntranceRateDirectFee: ChecksumAddress
    FundActionsWrapper: ChecksumAddress
    FundValueCalculator: ChecksumAddress
    GuaranteedRedemption: ChecksumAddress
    IdleAdapter: ChecksumAddress
    IdlePriceFeed: ChecksumAddress
    InvestorWhitelist: ChecksumAddress
    KyberAdapter: ChecksumAddress
    LidoStethPriceFeed: ChecksumAddress
    ManagementFee: ChecksumAddress
    MaxConcentration: ChecksumAddress
    MinMaxInvestment: ChecksumAddress
    ParaSwapV4Adapter: ChecksumAddress
    PerformanceFee: ChecksumAddress
    StakehoundEthPriceFeed: ChecksumAddress
    SynthetixAdapter: ChecksumAddress
    SynthetixPriceFeed: ChecksumAddress
    TrackedAssetsAdapter: ChecksumAddress
    UniswapV2Adapter: ChecksumAddress
    UniswapV2PoolPriceFeed: ChecksumAddress
    ValueInterpreter: ChecksumAddress
    WdgldPriceFeed: ChecksumAddress
    ZeroExV2Adapter: ChecksumAddress


Contracts = EncoreContracts | PhoenixContracts | SuluContracts


SuluContractNames = Literal[
    "ComptrollerLib",
    "Dispatcher",
    "FeeManager",
    "FundDataProviderRouter",
    "FundDeployer",
    "FundValueCalculatorRouter",
    "IntegrationManager",
    "PolicyManager",
    "VaultLib",
    "AaveDebtPositionLib",
    "AaveDebtPositionParser",
    "AaveV3DebtPositionLib",
    "AaveV3DebtPositionParser",
    "AaveV3FlashLoanAssetManagerFactory",
    "AaveV3FlashLoanAssetManagerLib",
    "AavePriceFeed",
    "AaveV2Adapter",
    "AaveV2ATokenListOwner",
    "AaveV3Adapter",
    "AaveV3ATokenListOwner",
    "AddressListRegistry",
    "AlicePositionLib",
    "AlicePositionParser",
    "AllowedAdapterIncomingAssetsPolicy",
    "AllowedAdaptersPerManagerPolicy",
    "AllowedAdaptersPolicy",
    "AllowedAssetsForRedemptionPolicy",
    "AllowedDepositRecipientsPolicy",
    "AllowedExternalPositionTypesPerManagerPolicy",
    "AllowedExternalPositionTypesPolicy",
    "AllowedRedeemersForSpecificAssetsPolicy",
    "AllowedSharesTransferRecipientsPolicy",
    "ArbitraryLoanPositionLib",
    "ArbitraryLoanPositionParser",
    "ArbitraryLoanTotalNominalDeltaOracleModule",
    "ArrakisV2Adapter",
    "ArrakisV2PriceFeed",
    "AssetValueCalculator",
    "BalancerV2GaugeTokenPriceFeed",
    "BalancerV2LiquidityAdapter",
    "BalancerV2StablePoolPriceFeed",
    "BalancerV2WeightedPoolPriceFeed",
    "ChainlinkLikeWstethPriceFeed",
    "ChainlinkLikeYnEthPriceFeed",
    "CompoundAdapter",
    "CompoundDebtPositionLib",
    "CompoundDebtPositionParser",
    "CompoundPriceFeed",
    "CompoundV3TokenListOwner",
    "CompoundV3Adapter",
    "ConvexVotingPositionLib",
    "ConvexVotingPositionParser",
    "CumulativeSlippageTolerancePolicy",
    "CurveExchangeAdapter",
    "CurveLiquidityAaveAdapter",
    "CurveLiquidityAdapter",
    "CurveLiquiditySethAdapter",
    "CurveLiquidityStethAdapter",
    "CurvePriceFeed",
    "DepositWrapper",
    "DisallowedAdapterIncomingAssetsPolicy",
    "EntranceRateBurnFee",
    "EntranceRateDirectFee",
    "ERC4626Adapter",
    "ERC4626PriceFeed",
    "EtherFiEthPriceFeed",
    "ExitRateBurnFee",
    "ExitRateDirectFee",
    "ExternalPositionFactory",
    "ExternalPositionManager",
    "FiduPriceFeed",
    "FundValueCalculator",
    "GasRelayPaymasterFactory",
    "GasRelayPaymasterLib",
    "GatedRedemptionQueueSharesWrapperFactory",
    "GatedRedemptionQueueSharesWrapperLib",
    "GenericAdapter",
    "GlobalConfigLib",
    "GlobalConfigProxy",
    "GMXV2LeverageTradingPositionLib",
    "GMXV2LeverageTradingPositionParser",
    "KilnStakingPositionLib",
    "KilnStakingPositionParser",
    "LidoWithdrawalsPositionLib",
    "LidoWithdrawalsPositionParser",
    "LiquityDebtPositionLib",
    "LiquityDebtPositionParser",
    "ManagementFee",
    "ManualValueOracleFactory",
    "MapleLiquidityPositionLib",
    "MapleLiquidityPositionParser",
    "MinAssetBalancesPostRedemptionPolicy",
    "MinMaxInvestmentPolicy",
    "MinSharesSupplyFee",
    "MorphoBluePositionLib",
    "MorphoBluePositionParser",
    "NoDepegOnRedeemSharesForSpecificAssetsPolicy",
    "NotionalV2PositionLib",
    "NotionalV2PositionParser",
    "OneInchV5Adapter",
    "OnlyRemoveDustExternalPositionPolicy",
    "OnlyUntrackDustOrPricelessAssetsPolicy",
    "ParaSwapV5Adapter",
    "PeggedDerivativesPriceFeed",
    "PendleV2Adapter",
    "PendleV2PositionLib",
    "PendleV2PositionParser",
    "PendleMarketsRegistry",
    "PerformanceFee",
    "PoolTogetherV4Adapter",
    "PoolTogetherV4PriceFeed",
    "ProtocolFeeReserveLib",
    "ProtocolFeeReserveProxy",
    "ProtocolFeeTracker",
    "SharePriceThrottledAssetManagerLib",
    "SharePriceThrottledAssetManagerFactory",
    "SharesSplitterFactory",
    "SingleAssetRedemptionQueueLib",
    "SingleAssetRedemptionQueueFactory",
    "SolvV2BondBuyerPositionLib",
    "SolvV2BondBuyerPositionParser",
    "SolvV2BondIssuerPositionLib",
    "SolvV2BondIssuerPositionParser",
    "StaderSDPriceFeed",
    "StaderStakingAdapter",
    "StaderWithdrawalsPositionLib",
    "StaderWithdrawalsPositionParser",
    "StakeWiseV3StakingPositionLib",
    "StakeWiseV3StakingPositionParser",
    "SwellStakingAdapter",
    "SynthetixAdapter",
    "TermFinanceV1LendingPositionLib",
    "TermFinanceV1LendingPositionParser",
    "TheGraphDelegationPositionLib",
    "TheGraphDelegationPositionParser",
    "ThreeOneThirdAdapter",
    "TransferAssetsAdapter",
    "UintListRegistry",
    "UniswapV2ExchangeAdapter",
    "UniswapV2LiquidityAdapter",
    "UniswapV2PoolPriceFeed",
    "UniswapV3Adapter",
    "UniswapV3LiquidityPositionLib",
    "UniswapV3LiquidityPositionParser",
    "UnpermissionedActionsWrapper",
    "UsdEthSimulatedAggregator",
    "ValueInterpreter",
    "WstethPriceFeed",
    "YearnVaultV2Adapter",
    "YearnVaultV2PriceFeed",
    "ZeroExV2Adapter",
    "ZeroExV4Adapter",
    "ZeroExV4AdapterPmm2Kyc",
]


EncoreContractNames = Literal[
    "ComptrollerLib",
    "Dispatcher",
    "FeeManager",
    "FundDataProviderRouter",
    "FundDeployer",
    "FundValueCalculatorRouter",
    "IntegrationManager",
    "PolicyManager",
    "VaultLib",
    "AaveAdapter",
    "AavePriceFeed",
    "AdapterBlacklist",
    "AdapterWhitelist",
    "AggregatedDerivativePriceFeed",
    "AlphaHomoraV1Adapter",
    "AlphaHomoraV1PriceFeed",
    "AssetBlacklist",
    "AssetValueCalculator",
    "AssetWhitelist",
    "BuySharesCallerWhitelist",
    "ChainlinkPriceFeed",
    "CompoundAdapter",
    "CompoundPriceFeed",
    "CurveExchangeAdapter",
    "CurveLiquidityAaveAdapter",
    "CurveLiquidityEursAdapter",
    "CurveLiquiditySethAdapter",
    "CurveLiquidityStethAdapter",
    "CurvePriceFeed",
    "EntranceRateBurnFee",
    "EntranceRateDirectFee",
    "FundActionsWrapper",
    "FundValueCalculator",
    "GuaranteedRedemption",
    "IdleAdapter",
    "IdlePriceFeed",
    "InvestorWhitelist",
    "KyberAdapter",
    "LidoStethPriceFeed",
    "ManagementFee",
    "MaxConcentration",
    "MinMaxInvestment",
    "ParaSwapV4Adapter",
    "PerformanceFee",
    "StakehoundEthPriceFeed",
    "SynthetixAdapter",
    "SynthetixPriceFeed",
    "TrackedAssetsAdapter",
    "UniswapV2Adapter",
    "UniswapV2PoolPriceFeed",
    "ValueInterpreter",
    "WdgldPriceFeed",
    "YearnVaultV2Adapter",
    "YearnVaultV2PriceFeed",
    "ZeroExV2Adapter",
]


PhoenixContractNames = Literal[
    "ComptrollerLib",
    "Dispatcher",
    "FeeManager",
    "FundDataProviderRouter",
    "FundDeployer",
    "FundValueCalculatorRouter",
    "IntegrationManager",
    "PolicyManager",
    "VaultLib",
    "AaveAdapter",
    "AavePriceFeed",
    "AdapterBlacklist",
    "AdapterWhitelist",
    "AggregatedDerivativePriceFeed",
    "AlphaHomoraV1Adapter",
    "AlphaHomoraV1PriceFeed",
    "AssetBlacklist",
    "AssetValueCalculator",
    "AssetWhitelist",
    "BuySharesCallerWhitelist",
    "ChaiAdapter",
    "ChainlinkPriceFeed",
    "ChaiPriceFeed",
    "CompoundAdapter",
    "CompoundPriceFeed",
    "CurveExchangeAdapter",
    "CurveLiquidityAaveAdapter",
    "CurveLiquiditySethAdapter",
    "CurveLiquidityStethAdapter",
    "CurvePriceFeed",
    "EntranceRateBurnFee",
    "EntranceRateDirectFee",
    "FundActionsWrapper",
    "FundValueCalculator",
    "GuaranteedRedemption",
    "IdleAdapter",
    "IdlePriceFeed",
    "InvestorWhitelist",
    "KyberAdapter",
    "LidoStethPriceFeed",
    "ManagementFee",
    "MaxConcentration",
    "MinMaxInvestment",
    "ParaSwapV4Adapter",
    "PerformanceFee",
    "StakehoundEthPriceFeed",
    "SynthetixAdapter",
    "SynthetixPriceFeed",
    "TrackedAssetsAdapter",
    "UniswapV2Adapter",
    "UniswapV2PoolPriceFeed",
    "ValueInterpreter",
    "WdgldPriceFeed",
    "ZeroExV2Adapter",
]


VersionContractNames = EncoreContractNames | PhoenixContractNames | SuluContractNames



from decimal import Decimal


def to_int(value: Decimal, decimals: int) -> int:
    return int(value * 10**decimals)


def to_bps(value: Decimal) -> int:
    return int(value * 10000)


def to_wei(value: Decimal, decimals: int = 18) -> int:
    return int(value * 10**decimals)


def from_wei(value: int, decimals: int = 18) -> Decimal:
    return Decimal(value) / Decimal(10**decimals)


def to_seconds(
    years: float = 0, weeks: float = 0, days: float = 0, hours: float = 0, minutes: float = 0
) -> int:
    return int(
        years * 31557600 + weeks * 604800 + days * 86400 + hours * 3600 + minutes * 60
    )



from web3.types import ChecksumAddress
from .utils.clients import PublicClient
from web3.exceptions import ContractLogicError
from ..abis import abis


async def get_nav(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
    value_calculator: ChecksumAddress,
) -> dict[ChecksumAddress, int]:
    contract = client.contract(value_calculator, abis.IFundValueCalculatorRouter)
    function = contract.functions.calcNav(vault_proxy)
    asset, value = await function.call()
    return {"asset": asset, "value": value}


async def get_nav_in_asset(
    client: PublicClient,
    asset: ChecksumAddress,
    vault_proxy: ChecksumAddress,
    value_calculator: ChecksumAddress,
) -> int:
    contract = client.contract(value_calculator, abis.IFundValueCalculatorRouter)
    function = contract.functions.calcNavInAsset(vault_proxy, asset)
    return await function.call()


async def get_gav(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
    value_calculator: ChecksumAddress,
) -> dict[ChecksumAddress, int]:
    contract = client.contract(value_calculator, abis.IFundValueCalculatorRouter)
    function = contract.functions.calcGav(vault_proxy)
    asset, value = await function.call()
    return {"asset": asset, "value": value}


async def get_gav_in_asset(
    client: PublicClient,
    asset: ChecksumAddress,
    vault_proxy: ChecksumAddress,
    value_calculator: ChecksumAddress,
) -> int:
    contract = client.contract(value_calculator, abis.IFundValueCalculatorRouter)
    function = contract.functions.calcGavInAsset(vault_proxy, asset)
    return await function.call()


async def get_share_price(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
    value_calculator: ChecksumAddress,
) -> dict[ChecksumAddress, int]:
    contract = client.contract(value_calculator, abis.IFundValueCalculatorRouter)
    function = contract.functions.calcNetShareValue(vault_proxy)
    asset, value = await function.call()
    return {"asset": asset, "value": value}


async def get_share_price_in_asset(
    client: PublicClient,
    asset: ChecksumAddress,
    vault_proxy: ChecksumAddress,
    value_calculator: ChecksumAddress,
) -> int:
    contract = client.contract(value_calculator, abis.IFundValueCalculatorRouter)
    function = contract.functions.calcNetShareValueInAsset(vault_proxy, asset)
    return await function.call()


async def get_canonical_asset_value(
    client: PublicClient,
    asset: ChecksumAddress,
    value_calculator: ChecksumAddress,
) -> int:
    contract = client.contract(value_calculator, abis.IValueInterpreter)
    function = contract.functions.calcCanonicalAssetValue(asset)
    try:
        return await function.call()
    except ContractLogicError:  # TODO: More selectively catch this error here.
        return None


async def calc_canonical_assets_total_value(
    client: PublicClient,
    value_interpreter: ChecksumAddress,
    base_assets: list[ChecksumAddress],
    amounts: list[int],
    quote_asset: ChecksumAddress,
) -> int:
    contract = client.contract(value_interpreter, abis.IValueInterpreter)
    function = contract.functions.calcCanonicalAssetsTotalValue(
        base_assets, amounts, quote_asset
    )
    try:
        return await function.call()
    except ContractLogicError:  # TODO: More selectively catch this error here.
        return None



from typing import TypedDict
from web3 import Web3
from web3.types import ChecksumAddress, HexStr, TxParams
from eth_abi import encode, decode
from ..._internal import integration_manager as integration_manager_lib
from ...utils.encoding import encoding_to_types

# --------------------------------------------------------------------------------------------
# TAKE ORDER
# --------------------------------------------------------------------------------------------


async def take_order(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        integration_manager: ChecksumAddress
        integration_adapter: ChecksumAddress
        call_args: dict[str, Any]
            {
                "executor": ChecksumAddress,
                "order_description": dict[str, Any],
                "data": HexStr,
            }
    """
    _take_order = integration_manager_lib.make_use(
        integration_manager_lib.SELECTOR["take_order"], take_order_encode
    )
    return await _take_order(*args)


TAKE_ORDER_ENCODING = [
    {
        "name": "executor",
        "type": "address",
    },
    {
        "components": [
            {
                "internalType": "address",
                "name": "srcToken",
                "type": "address",
            },
            {
                "internalType": "address",
                "name": "dstToken",
                "type": "address",
            },
            {
                "internalType": "address",
                "name": "srcReceiver",
                "type": "address",
            },
            {
                "internalType": "address",
                "name": "dstReceiver",
                "type": "address",
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256",
            },
            {
                "internalType": "uint256",
                "name": "minReturnAmount",
                "type": "uint256",
            },
            {
                "internalType": "uint256",
                "name": "flags",
                "type": "uint256",
            },
        ],
        "name": "orderDescription",
        "type": "tuple",
    },
    {
        "name": "data",
        "type": "bytes",
    },
]


class OrderDescription(TypedDict):
    srcToken: ChecksumAddress
    dstToken: ChecksumAddress
    srcReceiver: ChecksumAddress
    dstReceiver: ChecksumAddress
    amount: int
    minReturnAmount: int
    flags: int


class TakeOrderArgs(TypedDict):
    executor: ChecksumAddress
    order_description: OrderDescription
    data: HexStr


def take_order_encode(
    executor: ChecksumAddress,
    order_description: OrderDescription,
    data: HexStr,
) -> HexStr:
    """
    Args:
        order_description:
            {
                "srcToken": ChecksumAddress,
                "dstToken": ChecksumAddress,
                "srcReceiver": ChecksumAddress,
                "dstReceiver": ChecksumAddress,
                "amount": int,
                "minReturnAmount": int,
                "flags": int,
            }
    """
    types = encoding_to_types(TAKE_ORDER_ENCODING)
    values = [
        executor,
        (
            order_description["srcToken"],
            order_description["dstToken"],
            order_description["srcReceiver"],
            order_description["dstReceiver"],
            order_description["amount"],
            order_description["minReturnAmount"],
            order_description["flags"],
        ),
        Web3.to_bytes(hexstr=data),
    ]
    return Web3.to_hex(encode(types, values))


def take_order_decode(
    encoded: HexStr,
) -> dict[str, ChecksumAddress | dict[str, ChecksumAddress | int] | HexStr]:
    """
    Returns:
        {
            "executor": ChecksumAddress,
            "order_description": {
                "src_token": ChecksumAddress,
                "dst_token": ChecksumAddress,
                "src_receiver": ChecksumAddress,
                "dst_receiver": ChecksumAddress,
                "amount": int,
                "min_return_amount": int,
                "flags": int,
            },
            "data": HexStr,
        }
    """
    types = encoding_to_types(TAKE_ORDER_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "executor": Web3.to_checksum_address(decoded[0]),
        "order_description": {
            "src_token": Web3.to_checksum_address(decoded[1][0]),
            "dst_token": Web3.to_checksum_address(decoded[1][1]),
            "src_receiver": Web3.to_checksum_address(decoded[1][2]),
            "dst_receiver": Web3.to_checksum_address(decoded[1][3]),
            "amount": decoded[1][4],
            "min_return_amount": decoded[1][5],
            "flags": decoded[1][6],
        },
        "data": Web3.to_hex(decoded[2]),
    }


SWAP_ARGS_ENCODING = [
    {
        "name": "executor",
        "type": "address",
    },
    {
        "components": [
            {
                "internalType": "address",
                "name": "srcToken",
                "type": "address",
            },
            {
                "internalType": "address",
                "name": "dstToken",
                "type": "address",
            },
            {
                "internalType": "address",
                "name": "srcReceiver",
                "type": "address",
            },
            {
                "internalType": "address",
                "name": "dstReceiver",
                "type": "address",
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256",
            },
            {
                "internalType": "uint256",
                "name": "minReturnAmount",
                "type": "uint256",
            },
            {
                "internalType": "uint256",
                "name": "flags",
                "type": "uint256",
            },
        ],
        "name": "orderDescription",
        "type": "tuple",
    },
    {
        "name": "unknown",
        "type": "bytes",
    },
    {
        "name": "data",
        "type": "bytes",
    },
]


def decoded_swap_args(
    encoded: HexStr,
) -> dict[str, ChecksumAddress | dict[str, ChecksumAddress | int] | HexStr]:
    """
    Returns:
        {
            "executor": ChecksumAddress,
            "order_description": {
                "src_token": ChecksumAddress,
                "dst_token": ChecksumAddress,
                "src_receiver": ChecksumAddress,
                "dst_receiver": ChecksumAddress,
                "amount": int,
                "min_return_amount": int,
                "flags": int,
            },
            "data": HexStr,
        }
    """
    types = encoding_to_types(SWAP_ARGS_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "executor": Web3.to_checksum_address(decoded[0]),
        "order_description": {
            "src_token": Web3.to_checksum_address(decoded[1][0]),
            "dst_token": Web3.to_checksum_address(decoded[1][1]),
            "src_receiver": Web3.to_checksum_address(decoded[1][2]),
            "dst_receiver": Web3.to_checksum_address(decoded[1][3]),
            "amount": decoded[1][4],
            "min_return_amount": decoded[1][5],
            "flags": decoded[1][6],
        },
        "data": Web3.to_hex(decoded[3]),
    }



from typing import Literal, TypedDict
from web3.types import ChecksumAddress, HexStr
from .price_feeds import PriceFeed
from .types import Network, Release


AssetType = Literal[
    "aave-v2"
    "aave-v3"
    "balancer-pool"
    "balancer-pool-gauge"
    "compound-v2"
    "compound-v3"
    "curve-pool-lp"
    "curve-pool-gauge"
    "idle"
    "pendle-v2-lp"
    "pendle-v2-pt"
    "primitive"
    "stader"
    "synthetix"
    "uniswap-v2-pool"
    "yearn-vault-v2"
    "maple-v1"
    "maple-v2"
    "erc-4626"
    "zero-lend-lrt-btc-aave-v3"
    "zero-lend-rwa-stablecoins-aave-v3"
]


class AssetDefinitionInput(TypedDict):
    # The asset address
    id: ChecksumAddress
    # The asset name
    name: str
    # The asset symbol
    symbol: str
    # Number of decimal places
    decimals: int
    # The asset type
    type: AssetType
    # List of release slugs the asset is registered on
    releases: list[Release]
    # Price feed used for the asset
    price_feed: PriceFeed


class AssetBase(AssetDefinitionInput):
    # The network the asset is deployed on
    network: Network
    # Whether the asset is registered on the current release
    registered: bool


class StaderAsset(AssetBase):
    type: Literal["stader"]


class SynthetixAsset(AssetBase):
    type: Literal["synthetix"]


class PrimitiveAsset(AssetBase):
    type: Literal["primitive"]


class AaveV2Asset(AssetBase):
    type: Literal["aave-v2"]
    # Underlying Asset
    underlying: ChecksumAddress


class AaveV3Asset(AssetBase):
    type: Literal["aave-v3"]
    # Underlying Asset
    underlying: ChecksumAddress


class ZeroLendLRTBTCAaveV3Asset(AssetBase):
    type: Literal["zero-lend-lrt-btc-aave-v3"]
    # Underlying Asset
    underlying: ChecksumAddress


class ZeroLendRWAStablecoinsAaveV3Asset(AssetBase):
    type: Literal["zero-lend-rwa-stablecoins-aave-v3"]
    # Underlying Asset
    underlying: ChecksumAddress


class CompoundV2Asset(AssetBase):
    type: Literal["compound-v2"]
    # Underlying Asset
    underlying: ChecksumAddress


class CompoundV3Asset(AssetBase):
    type: Literal["compound-v3"]
    # Underlying Asset
    underlying: ChecksumAddress


class IdleAsset(AssetBase):
    type: Literal["idle"]
    # Underlying Asset
    underlying: ChecksumAddress


class YearnVaultV2Asset(AssetBase):
    type: Literal["yearn-vault-v2"]
    # Underlying Asset
    underlying: ChecksumAddress


class UniswapV2PoolAsset(AssetBase):
    type: Literal["uniswap-v2-pool"]
    # Underlying Asset Pair
    underlyings: list[ChecksumAddress]


class MapleV1Asset(AssetBase):
    type: Literal["maple-v1"]
    # Underlying Asset
    underlying: ChecksumAddress
    rewards_contract: ChecksumAddress


class MapleV2Asset(AssetBase):
    type: Literal["maple-v2"]
    # Underlying Asset
    underlying: ChecksumAddress
    pool_manager: ChecksumAddress


Erc4626Protocol = Literal["angle", "morpho", "sky", "spark", "vaultcraft"]


class ERC4626Asset(AssetBase):
    type: Literal["erc-4626"]
    # The protocol of the ERC4626 asset (since there are multiple protocols that implement ERC4626 assets)
    protocol: Erc4626Protocol
    # Underlying Asset
    underlying: ChecksumAddress


BalancerStakingType = Literal["none", "gauge"]


class BalancerStaking(TypedDict):
    token: ChecksumAddress
    type: BalancerStakingType


BalancerPoolType = Literal["meta-stable", "stable", "weighted", "composable-stable-v5"]


class BalancerPoolAsset(AssetBase):
    type: Literal["balancer-pool"]
    # Balancer pool factory
    pool_factory: ChecksumAddress
    # Staking options
    staking: list[BalancerStaking]
    # Balancer pool type
    pool_type: BalancerPoolType
    # Underlying Assets
    underlyings: list[ChecksumAddress]
    # Pool id
    pool_id: HexStr


class BalancerPoolGaugeAsset(AssetBase):
    type: Literal["balancer-pool-gauge"]
    # Balancer Pool
    pool: ChecksumAddress
    # Balancer gauge factory
    gauge_factory: ChecksumAddress
    # Balancer pool template
    pool_type: BalancerPoolType
    # Underlying Assets
    underlyings: list[ChecksumAddress]


CurvePoolTemplate = Literal["aave", "base", "eth", "meta", "yearn"]


CurveStakingType = Literal["none", "gauge"]


class CurveStaking(TypedDict):
    token: ChecksumAddress
    type: CurveStakingType


class CurvePoolLpAsset(AssetBase):
    type: Literal["curve-pool-lp"]
    # Curve Pool
    pool: ChecksumAddress
    # Staking options
    staking: list[CurveStaking]
    # Curve pool template. See: https://github.com/curvefi/curve-contract/tree/master/contracts/pool-templates
    template: CurvePoolTemplate
    # Underlying Assets
    underlyings: list[ChecksumAddress]


class CurvePoolGaugeAsset(AssetBase):
    type: Literal["curve-pool-gauge"]
    # Curve LP Token
    lp: ChecksumAddress
    # Curve Pool
    pool: ChecksumAddress
    # Curve pool template. See: https://github.com/curvefi/curve-contract/tree/master/contracts/pool-templates
    template: CurvePoolTemplate
    # Underlying Assets
    underlyings: list[ChecksumAddress]


class PendleV2LPAsset(AssetBase):
    type: Literal["pendle-v2-lp"]
    # Underlying Asset
    underlying: ChecksumAddress


class PendleV2PTAsset(AssetBase):
    type: Literal["pendle-v2-pt"]
    # Underlying Asset
    underlying: ChecksumAddress
    # Markets
    markets: list[ChecksumAddress]


Asset = (
    AaveV2Asset
    | AaveV3Asset
    | BalancerPoolAsset
    | BalancerPoolGaugeAsset
    | CompoundV2Asset
    | CompoundV3Asset
    | CurvePoolGaugeAsset
    | CurvePoolLpAsset
    | ERC4626Asset
    | IdleAsset
    | MapleV1Asset
    | MapleV2Asset
    | PendleV2PTAsset
    | PendleV2LPAsset
    | PrimitiveAsset
    | StaderAsset
    | SynthetixAsset
    | UniswapV2PoolAsset
    | YearnVaultV2Asset
    | ZeroLendLRTBTCAaveV3Asset
    | ZeroLendRWAStablecoinsAaveV3Asset
)


def define_asset_list(
    network: Network, assets: list[AssetDefinitionInput]
) -> list[AssetBase]:
    return [asset | {"network": network, "registered": False} for asset in assets]



from enum import IntEnum
from web3.types import ChecksumAddress
from typing import Literal, TypedDict

PriceFeedType = Literal[
    "NONE",
    "WETH",
    "PRIMITIVE_CHAINLINK",
    "PRIMITIVE_CHAINLINK_LIKE_ETHX",
    "PRIMITIVE_CHAINLINK_LIKE_WSTETH",
    "PRIMITIVE_CHAINLINK_LIKE_YNETH",
    "PRIMITIVE_REDSTONE",
    "PRIMITIVE_REDSTONE_QUOTED",
    "PRIMITIVE_NON_STANDARD_PRECISION",
    "PRIMITIVE_PENDLE_V2",
    "DERIVATIVE_ARRAKIS_V2",
    "DERIVATIVE_BALANCER_V2_GAUGE_TOKEN",
    "DERIVATIVE_BALANCER_V2_STABLE_POOL",
    "DERIVATIVE_BALANCER_V2_WEIGHTED_POOL",
    "DERIVATIVE_COMPOUND",
    "DERIVATIVE_CURVE",
    "DERIVATIVE_ERC4626",
    "DERIVATIVE_ETHERFI",
    "DERIVATIVE_PEGGED_DERIVATIVES",
    "DERIVATIVE_REVERTING",
    "DERIVATIVE_STADER_SD",
    "DERIVATIVE_UNISWAP_V2_POOL",
    "DERIVATIVE_WSTETH",
    "DERIVATIVE_YEARN_VAULT_V2",
]


PRIMITIVE_PRICE_FEEDS = [
    "PRIMITIVE_CHAINLINK",
    "PRIMITIVE_CHAINLINK_LIKE_ETHX",
    "PRIMITIVE_CHAINLINK_LIKE_WSTETH",
    "PRIMITIVE_CHAINLINK_LIKE_YNETH",
    "PRIMITIVE_REDSTONE",
    "PRIMITIVE_REDSTONE_QUOTED",
    "PRIMITIVE_NON_STANDARD_PRECISION",
    "PRIMITIVE_PENDLE_V2",
]


DERIVATIVE_PRICE_FEEDS = [
    "DERIVATIVE_ARRAKIS_V2",
    "DERIVATIVE_BALANCER_V2_GAUGE_TOKEN",
    "DERIVATIVE_BALANCER_V2_STABLE_POOL",
    "DERIVATIVE_BALANCER_V2_WEIGHTED_POOL",
    "DERIVATIVE_COMPOUND",
    "DERIVATIVE_CURVE",
    "DERIVATIVE_ERC4626",
    "DERIVATIVE_ETHERFI",
    "DERIVATIVE_PEGGED_DERIVATIVES",
    "DERIVATIVE_STADER_SD",
    "DERIVATIVE_UNISWAP_V2_POOL",
    "DERIVATIVE_WSTETH",
    "DERIVATIVE_YEARN_VAULT_V2",
]


class RateAsset(IntEnum):
    ETH = 0
    USD = 1


class PriceFeedBase(TypedDict, total=False):
    non_standard: bool
    pegged_to: str
    comment: str


class NoPriceFeed(PriceFeedBase):
    type: Literal["NONE"]


class WethPriceFeed(PriceFeedBase):
    type: Literal["WETH"]


class PrimitiveChainlinkPriceFeed(PriceFeedBase):
    type: Literal["PRIMITIVE_CHAINLINK"]
    # Aggregator address
    aggregator: ChecksumAddress
    # Rate asset (ETH = 0, USD = 1)
    rate_asset: RateAsset


class PrimitiveChainlinkLikeWstEthPriceFeed(PriceFeedBase):
    type: Literal["PRIMITIVE_CHAINLINK_LIKE_WSTETH"]
    # Aggregator address
    aggregator: ChecksumAddress
    # Rate asset (ETH = 0, USD = 1)
    rate_asset: RateAsset


class PrimitiveChainlinkLikeYnEthPriceFeed(PriceFeedBase):
    type: Literal["PRIMITIVE_CHAINLINK_LIKE_YNETH"]
    # Aggregator address
    aggregator: ChecksumAddress
    # Rate asset (ETH = 0, USD = 1)
    rate_asset: RateAsset


class PrimitiveChainlinkLikeEthxPriceFeed(PriceFeedBase):
    type: Literal["PRIMITIVE_CHAINLINK_LIKE_ETHX"]
    # Aggregator address
    aggregator: ChecksumAddress
    # Rate asset (ETH = 0, USD = 1)
    rate_asset: RateAsset


class PrimitiveRedstonePriceFeed(PriceFeedBase):
    type: Literal["PRIMITIVE_REDSTONE"]
    # Aggregator address
    aggregator: ChecksumAddress
    # Rate asset (ETH = 0, USD = 1)
    rate_asset: RateAsset


class PrimitiveRedstoneQuotedPriceFeed(PriceFeedBase):
    type: Literal["PRIMITIVE_REDSTONE_QUOTED"]
    # Aggregator address
    aggregator: ChecksumAddress
    # Rate asset (ETH = 0, USD = 1)
    rate_asset: RateAsset


class PrimitiveNonStandardPrecisionPriceFeed(PriceFeedBase):
    type: Literal["PRIMITIVE_REDSTONE_NON_STANDARD_PRECISION"]
    # Aggregator address
    aggregator: ChecksumAddress
    # Rate asset (ETH = 0, USD = 1)
    rate_asset: RateAsset


class PrimitivePendleV2PriceFeed(PriceFeedBase):
    type: Literal["PRIMITIVE_PENDLE_V2"]
    # Aggregator address
    aggregator: ChecksumAddress
    # Rate asset (ETH = 0, USD = 1)
    rate_asset: RateAsset


class DerivativeArrakisV2PriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_ARRAKIS_V2"]
    # Price feed address
    address: ChecksumAddress


class DerivativeBalancerV2GaugeTokenPriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_BALANCER_V2_GAUGE_TOKEN"]
    # Price feed address
    address: ChecksumAddress


class DerivativeBalancerV2StablePoolPriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_BALANCER_V2_STABLE_POOL"]
    # Price feed address
    address: ChecksumAddress
    # Invariant Proxy Asset (ipa)
    ipa: ChecksumAddress


class DerivativeBalancerV2WeightedPoolPriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_BALANCER_V2_WEIGHTED_POOL"]
    # Price feed address
    address: ChecksumAddress


class DerivativeCompoundPriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_COMPOUND"]
    # Price feed address
    address: ChecksumAddress


class DerivativeCurvePriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_CURVE"]
    # Price feed address
    address: ChecksumAddress
    # Invariant Proxy Asset (ipa)
    ipa: ChecksumAddress


class DerivativeERC4626PriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_ERC4626"]
    # Price feed address
    address: ChecksumAddress


class DerivativeEtherfiPriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_ETHERFI"]
    # Price feed address
    address: ChecksumAddress


class DerivativePeggedDerivativesPriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_PEGGED_DERIVATIVES"]
    # Price feed address
    address: ChecksumAddress


class DerivativeStaderSDPriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_STADER_SD"]
    # Price feed address
    address: ChecksumAddress


class DerivativeUniswapV2PoolPriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_UNISWAP_V2_POOL"]
    # Price feed address
    address: ChecksumAddress


class DerivativeWstethPriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_WSTETH"]
    # Price feed address
    address: ChecksumAddress


class DerivativeYearnVaultV2PriceFeed(PriceFeedBase):
    type: Literal["DERIVATIVE_YEARN_VAULT_V2"]
    # Price feed address
    address: ChecksumAddress


PriceFeed = (
  NoPriceFeed
  | WethPriceFeed
  | PrimitiveChainlinkPriceFeed
  | PrimitiveChainlinkLikeEthxPriceFeed
  | PrimitiveChainlinkLikeWstEthPriceFeed
  | PrimitiveChainlinkLikeYnEthPriceFeed
  | PrimitiveRedstonePriceFeed
  | PrimitiveRedstoneQuotedPriceFeed
  | PrimitiveNonStandardPrecisionPriceFeed
  | PrimitivePendleV2PriceFeed
  | DerivativeArrakisV2PriceFeed
  | DerivativeBalancerV2GaugeTokenPriceFeed
  | DerivativeBalancerV2StablePoolPriceFeed
  | DerivativeBalancerV2WeightedPoolPriceFeed
  | DerivativeCompoundPriceFeed
  | DerivativeCurvePriceFeed
  | DerivativeERC4626PriceFeed
  | DerivativeEtherfiPriceFeed
  | DerivativePeggedDerivativesPriceFeed
  | DerivativeStaderSDPriceFeed
  | DerivativeUniswapV2PoolPriceFeed
  | DerivativeWstethPriceFeed
  | DerivativeYearnVaultV2PriceFeed
)



from web3 import Web3
from web3.types import ChecksumAddress, HexStr, TxParams
from eth_abi import encode, decode
from typing import Callable, Any, TypedDict
from .extensions import call_extension
from ..utils.clients import WalletClient
from ..utils.encoding import encoding_to_types


class Action(TypedDict):
    call_on_integration: int
    add_tracked_assets: int
    remove_tracked_assets: int


ACTION = {
    "call_on_integration": 0,
    "add_tracked_assets": 1,
    "remove_tracked_assets": 2,
}


class Selector(TypedDict):
    action: HexStr
    claim_rewards: HexStr
    lend: HexStr
    lend_and_stake: HexStr
    redeem: HexStr
    stake: HexStr
    take_multiple_orders: HexStr
    take_order: HexStr
    transfer: HexStr
    unstake: HexStr
    unstake_and_redeem: HexStr
    wrap: HexStr


SELECTOR = {
    "action": "0xa7a19e00",  # action(address,bytes,bytes)
    "claim_rewards": "0xb9dfbacc",  # claimRewards(address,bytes,bytes)
    "lend": "0x099f7515",  # lend(address,bytes,bytes)
    "lend_and_stake": "0x29fa046e",  # lendAndStake(address,bytes,bytes)
    "redeem": "0xc29fa9dd",  # redeem(address,bytes,bytes)
    "stake": "0xfa7dd04d",  # stake(address,bytes,bytes)
    "take_multiple_orders": "0x0e7f692d",  # takeMultipleOrders(address,bytes,bytes)
    "take_order": "0x03e38a2b",  # takeOrder(address,bytes,bytes)
    "transfer": "0x3461917c",  # transfer(address,bytes,bytes)
    "unstake": "0x68e30677",  # unstake(address,bytes,bytes)
    "unstake_and_redeem": "0x8334eb99",  # unstakeAndRedeem(address,bytes,bytes)
    "wrap": "0xa5ca2d71",  # wrap(address,bytes,bytes)
}


class UseParams(TypedDict):
    client: WalletClient
    comptroller_proxy: ChecksumAddress
    integration_manager: ChecksumAddress
    integration_adapter: ChecksumAddress
    call_args: dict[str, Any]


def make_use(selector: HexStr, encoder: Callable) -> Callable:
    async def use_integration(
        client: WalletClient,
        comptroller_proxy: ChecksumAddress,
        integration_manager: ChecksumAddress,
        integration_adapter: ChecksumAddress,
        call_args: dict[str, Any],
    ) -> TxParams:
        return await call(
            client,
            comptroller_proxy,
            integration_manager,
            integration_adapter,
            selector,
            encoder(**call_args),
        )

    return use_integration


# --------------------------------------------------------------------------------------------
# CALL ON INTEGRATION
# --------------------------------------------------------------------------------------------

CALL_ENCODING = [
    {
        "type": "address",
        "name": "adapter",
    },
    {
        "type": "bytes4",
        "name": "selector",
    },
    {
        "type": "bytes",
        "name": "integrationData",
    },
]


class CallArgs(TypedDict):
    function_selector: HexStr
    integration_adapter: ChecksumAddress
    call_args: HexStr | None


def call_encode(
    function_selector: HexStr,
    integration_adapter: ChecksumAddress,
    call_args: HexStr | None,
) -> HexStr:
    types = encoding_to_types(CALL_ENCODING)
    values = [
        integration_adapter,
        Web3.to_bytes(hexstr=function_selector),
        Web3.to_bytes(hexstr=call_args or "0x"),
    ]
    return Web3.to_hex(encode(types, values))


def call_decode(encoded: HexStr) -> dict[str, HexStr | ChecksumAddress | HexStr]:
    """
    Returns:
            {
            "function_selector": HexStr,
            "integration_adapter": ChecksumAddress,
            "call_args": HexStr,
        }
    """
    types = encoding_to_types(CALL_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "function_selector": Web3.to_hex(decoded[0]),
        "integration_adapter": decoded[1],
        "call_args": Web3.to_hex(decoded[2]),
    }


class CallParams(TypedDict):
    client: WalletClient
    comptroller_proxy: ChecksumAddress
    integration_manager: ChecksumAddress
    integration_adapter: ChecksumAddress
    function_selector: HexStr
    call_args: HexStr | None


async def call(
    client: WalletClient,
    comptroller_proxy: ChecksumAddress,
    integration_manager: ChecksumAddress,
    integration_adapter: ChecksumAddress,
    function_selector: HexStr,
    call_args: HexStr | None,
) -> TxParams:
    return await call_extension(
        client,
        comptroller_proxy,
        integration_manager,
        ACTION["call_on_integration"],
        call_encode(function_selector, integration_adapter, call_args or "0x"),
    )


# --------------------------------------------------------------------------------------------
# ADD TRACKED ASSET
# --------------------------------------------------------------------------------------------

ADD_TRACKED_ASSETS_ENCODING = [
    {
        "type": "address[]",
        "name": "assets",
    },
]


class AddTrackedAssetsArgs(TypedDict):
    add_assets: list[ChecksumAddress]


def add_tracked_assets_encode(add_assets: list[ChecksumAddress]) -> HexStr:
    types = encoding_to_types(ADD_TRACKED_ASSETS_ENCODING)
    values = [add_assets]
    return Web3.to_hex(encode(types, values))


def add_tracked_assets_decode(encoded: HexStr) -> dict[str, list[ChecksumAddress]]:
    """
    Returns:
        {
            "assets": list[ChecksumAddress],
        }
    """
    types = encoding_to_types(ADD_TRACKED_ASSETS_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "assets": decoded[0],
    }


class AddTrackedAssetsParams(TypedDict):
    client: WalletClient
    comptroller_proxy: ChecksumAddress
    integration_manager: ChecksumAddress
    add_assets: list[ChecksumAddress]


async def add_tracked_assets(
    client: WalletClient,
    comptroller_proxy: ChecksumAddress,
    integration_manager: ChecksumAddress,
    add_assets: list[ChecksumAddress],
) -> TxParams:
    return await call_extension(
        client,
        comptroller_proxy,
        integration_manager,
        ACTION["add_tracked_assets"],
        add_tracked_assets_encode(add_assets),
    )


# --------------------------------------------------------------------------------------------
# REMOVE TRACKED ASSET
# --------------------------------------------------------------------------------------------

REMOVE_TRACKED_ASSETS_ENCODING = [
    {
        "type": "address[]",
        "name": "assets",
    },
]


class RemoveTrackedAssetsArgs(TypedDict):
    remove_assets: list[ChecksumAddress]


def remove_tracked_assets_encode(remove_assets: list[ChecksumAddress]) -> HexStr:
    types = encoding_to_types(REMOVE_TRACKED_ASSETS_ENCODING)
    values = [remove_assets]
    return Web3.to_hex(encode(types, values))


def remove_tracked_assets_decode(encoded: HexStr) -> dict[str, list[ChecksumAddress]]:
    """
    Returns:
        {
            "assets": list[ChecksumAddress],
        }
    """
    types = encoding_to_types(REMOVE_TRACKED_ASSETS_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "assets": decoded[0],
    }


class RemoveTrackedAssetsParams(TypedDict):
    client: WalletClient
    comptroller_proxy: ChecksumAddress
    integration_manager: ChecksumAddress
    remove_assets: list[ChecksumAddress]


async def remove_tracked_assets(
    client: WalletClient,
    comptroller_proxy: ChecksumAddress,
    integration_manager: ChecksumAddress,
    remove_assets: list[ChecksumAddress],
) -> TxParams:
    return await call_extension(
        client,
        comptroller_proxy,
        integration_manager,
        ACTION["remove_tracked_assets"],
        remove_tracked_assets_encode(remove_assets),
    )



from typing import Literal


Protocol = Literal[
    "AAVE",
    "AAVE_V3",
    "AURA",
    "ALICE",
    "BALANCER",
    "COMPOUND_V2",
    "COMPOUND_V3",
    "CONVEX",
    "CURVE",
    "KILN",
    "LIQUITY",
    "MAPLE",
    "MORPHO",
    "ONE_INCH_V5",
    "SPARK",
    "THE_GRAPH",
    "UNISWAP",
    "UNISWAP_V2",
    "VAULTCRAFT",
    "YEARN_VAULT_V2",
]



from .networks import Network, NetworkSlug, is_network_identifier, get_network
from .releases import Deployment, is_deployment
from .deployments_lib.all import DEPLOYMENTS


def get_deployment(
    deployment_or_network: Deployment | Network | NetworkSlug,
) -> Deployment:
    if is_deployment(deployment_or_network):
        return DEPLOYMENTS[deployment_or_network]
    if is_network_identifier(deployment_or_network):
        network = get_network(deployment_or_network)
        for deployment in DEPLOYMENTS.values():
            if deployment["network"] == network["id"] and deployment["kind"] == "live":
                return deployment
        raise ValueError(f"Missing deployment for network {network['slug']}")
    raise ValueError(f"Invalid deployment or network {deployment_or_network}")



import asyncio
from typing import TypedDict
from web3.types import ChecksumAddress, TxParams
from .utils.clients import WalletClient, PublicClient
from ..abis import abis


# --------------------------------------------------------------------------------------------
# TRANSACTIONS
# --------------------------------------------------------------------------------------------


class ApproveParams(TypedDict):
    client: WalletClient
    asset: ChecksumAddress
    spender: ChecksumAddress
    amount: int


async def approve(
    client: WalletClient,
    asset: ChecksumAddress,
    spender: ChecksumAddress,
    amount: int,
) -> TxParams:
    contract = client.contract(asset, abis.IERC20)
    function = contract.functions.approve(spender, amount)
    return await function.call()


# --------------------------------------------------------------------------------------------
# READ FUNCTIONS
# --------------------------------------------------------------------------------------------


async def get_info(
    client: PublicClient,
    asset: ChecksumAddress,
) -> dict[str, str | int]:
    """
    Returns:
        {
            "name": str,
            "symbol": str,
            "decimals": int,
        }
    """
    info = await asyncio.gather(
        get_name(client, asset),
        get_symbol(client, asset),
        get_decimals(client, asset),
    )
    return {
        "name": info[0],
        "symbol": info[1],
        "decimals": info[2],
    }


async def get_name(
    client: PublicClient,
    asset: ChecksumAddress,
) -> str:
    # TODO: Handle case where name is a bytes32
    contract = client.contract(asset, abis.IERC20)
    function = contract.functions.name()
    return await function.call()


async def get_symbol(
    client: PublicClient,
    asset: ChecksumAddress,
) -> str:
    # TODO: Handle case where symbol is a bytes32
    contract = client.contract(asset, abis.IERC20)
    function = contract.functions.symbol()
    return await function.call()


async def get_balance_of(
    client: PublicClient,
    owner: ChecksumAddress,
    asset: ChecksumAddress,
) -> int:
    contract = client.contract(asset, abis.IERC20)
    function = contract.functions.balanceOf(owner)
    return await function.call()


async def get_balances_of(
    client: PublicClient,
    owner: ChecksumAddress,
    assets: list[ChecksumAddress],
) -> list[dict[ChecksumAddress, int]]:
    """
    Returns:
        [
            {
                "asset": ChecksumAddress,
                "balance": int,
            },
            ...
        ]
    """
    balances = await asyncio.gather(
        *[get_balance_of(client, owner, asset) for asset in assets],
    )
    return [
        {
            "asset": assets[i],
            "balance": balances[i],
        }
        for i in range(len(assets))
    ]


async def get_allowance(
    client: PublicClient,
    asset: ChecksumAddress,
    owner: ChecksumAddress,
    spender: ChecksumAddress,
) -> int:
    contract = client.contract(asset, abis.IERC20)
    function = contract.functions.allowance(owner, spender)
    return await function.call()


async def get_decimals(
    client: PublicClient,
    asset: ChecksumAddress,
) -> int:
    contract = client.contract(asset, abis.IERC20)
    function = contract.functions.decimals()
    return await function.call()


async def get_total_supply(
    client: PublicClient,
    asset: ChecksumAddress,
) -> int:
    contract = client.contract(asset, abis.IERC20)
    function = contract.functions.totalSupply()
    return await function.call()


async def get_canonical_value(
    client: PublicClient,
    value_interpreter: ChecksumAddress,
    base_asset: ChecksumAddress,
    quote_asset: ChecksumAddress,
    amount: int,
) -> int:
    contract = client.contract(value_interpreter, abis.IValueInterpreter)
    function = contract.functions.calcCanonicalAssetValue(
        base_asset, amount, quote_asset
    )
    return await function.call()
